from solaredge2mqtt.eventbus import EventBus
from solaredge2mqtt.exceptions import InvalidDataException
from solaredge2mqtt.logging import logger
from solaredge2mqtt.models.homeassistant import HomeAssistantDevice, HomeAssistantEntity
from solaredge2mqtt.mqtt import MQTTPublishEvent
from solaredge2mqtt.service.modbus import Modbus
from solaredge2mqtt.settings import ServiceSettings


class HomeAssistantDiscovery:
    def __init__(self, service_settings: ServiceSettings, event_bus: EventBus) -> None:
        self.settings = service_settings
        self.event_bus = event_bus

    async def publish_discovery(self) -> None:
        modbus = Modbus(self.settings.modbus)
        inv_data, meters_data, batteries_data = await modbus.loop()

        if any(data is None for data in [inv_data, meters_data, batteries_data]):
            raise InvalidDataException("Invalid modbus data")

        device_info = inv_data.homeassistant_device_info()
        logger.trace(device_info)
        device = HomeAssistantDevice(
            client_id=self.settings.mqtt.client_id,
            state_topic=f"{self.settings.mqtt.topic_prefix}/{inv_data.mqtt_topic()}",
            **device_info,
        )
        logger.debug(device)

        entities_info = inv_data.homeassistant_entities_info()
        logger.trace(entities_info)
        for entity_info in entities_info:
            entity_type = entity_info.pop("type")
            if entity_type != "sensor":
                continue

            entity = HomeAssistantEntity(
                device=device,
                **entity_info,
            )
            topic = f"{entity_type}/{entity.unique_id}/config"
            logger.debug(entity)

            await MQTTPublishEvent.emit(
                self.event_bus,
                topic=topic,
                payload=entity,
                topic_prefix=self.settings.homeassistant.topic_prefix,
                exclude_none=True,
            )



"""
    This module, service.py, is part of the SolarEdge2MQTT service, which reads data 
    from a SolarEdge inverter and publishes it to an MQTT broker. It uses the asyncio 
    library for asynchronous I/O and the aiomqtt library for MQTT communication. 
    The module also includes a run function to initialize and start the service.
"""

import asyncio as aio
import signal
import platform
from time import time
from typing import Callable

from aiomqtt import MqttError

from solaredge2mqtt import __version__
from solaredge2mqtt.eventbus import EventBus
from solaredge2mqtt.exceptions import ConfigurationException, InvalidDataException
from solaredge2mqtt.logging import initialize_logging, logger
from solaredge2mqtt.mqtt import MQTTClient
from solaredge2mqtt.service.base import BaseLoops
from solaredge2mqtt.service.forecast import Forecast
from solaredge2mqtt.service.homeassistant import HomeAssistantDiscovery
from solaredge2mqtt.service.influxdb import InfluxDB
from solaredge2mqtt.service.monitoring import MonitoringSite
from solaredge2mqtt.service.weather import WeatherClient
from solaredge2mqtt.settings import LOCAL_TZ, service_settings


def run():
    try:
        service = Service()
        loop = aio.get_event_loop()
        loop.add_signal_handler(signal.SIGINT, service.cancel)
        loop.add_signal_handler(signal.SIGTERM, service.cancel)
        loop.run_until_complete(service.main_loop())
        loop.close()
    except ConfigurationException:
        logger.error("Configuration error")


class Service:
    def __init__(self):
        self.settings = service_settings()
        initialize_logging(self.settings.logging_level)
        logger.debug(self.settings)

        self.event_bus = EventBus()

        self.mqtt = MQTTClient(self.settings.mqtt, self.event_bus)

        self.cancel_request = aio.Event()
        self.loops: set[aio.Task] = set()

        self.influxdb: InfluxDB | None = (
            InfluxDB(self.settings.influxdb, self.settings.prices)
            if self.settings.is_influxdb_configured
            else None
        )

        self.basics = BaseLoops(self.settings, self.event_bus, self.influxdb)

        self.monitoring: MonitoringSite | None = (
            MonitoringSite(self.settings.monitoring, self.event_bus)
            if self.settings.is_monitoring_configured
            else None
        )

        self.weather: WeatherClient | None = (
            WeatherClient(self.settings, self.event_bus)
            if self.settings.is_weather_configured
            else None
        )

        self.forecast: Forecast | None = (
            Forecast(
                self.settings.forecast,
                self.settings.location,
                self.event_bus,
                self.influxdb,
            )
            if self.settings.is_forecast_configured
            else None
        )

        self.homeassistant: HomeAssistantDiscovery | None = (
            HomeAssistantDiscovery(self.settings, self.event_bus)
            if self.settings.is_homeassistant_configured
            else None
        )

    def cancel(self):
        logger.info("Stopping SolarEdge2MQTT service...")
        self.cancel_request.set()
        for loop in self.loops:
            loop.cancel()

    async def main_loop(self):
        logger.info("Starting SolarEdge2MQTT service...")
        logger.info("Version: {version}", version=__version__)
        logger.info(
            f"Operationg system: {platform.platform()} ({platform.system()}/{platform.machine()})"
        )
        logger.debug(self.settings)
        logger.info("Timezone: {timezone}", timezone=LOCAL_TZ)

        if self.settings.is_influxdb_configured:
            self.influxdb.initialize_buckets()

        while not self.cancel_request.is_set():
            try:
                async with self.mqtt:
                    await self.mqtt.publish_status_online()

                    if self.settings.is_homeassistant_configured:
                        logger.info("Home Assistant discovery enabled")
                        await self.homeassistant.publish_discovery()

                    self.schedule_loop(
                        self.settings.interval, self.basics.powerflow_loop
                    )

                    await self.schedule_weather_loops()

                    self.schedule_influxdb_loops()

                    self.schedule_monitoring_loop()

                    await aio.gather(*self.loops)

                    await self.mqtt.publish_status_offline()

            except MqttError:
                logger.error("MQTT error, reconnecting in 5 seconds...")
                await aio.sleep(5)
            except aio.exceptions.CancelledError:
                logger.debug("Loops cancelled")
                return
            finally:
                for loop in self.loops:
                    loop.cancel()

    def schedule_monitoring_loop(self):
        if self.settings.is_monitoring_configured:
            self.monitoring.login()
            self.schedule_loop(300, self.monitoring.loop)

    def schedule_influxdb_loops(self):
        if self.settings.is_influxdb_configured:
            loop_handles = [self.influxdb.loop, self.basics.energy_loop]
            self.schedule_loop(600, loop_handles)

    async def schedule_weather_loops(self):
        if self.settings.is_weather_configured:
            self.schedule_loop(600, self.weather.loop)

            if self.settings.is_forecast_configured:
                self.schedule_loop(600, self.forecast.forecast_loop, 300)

    def schedule_loop(
        self,
        interval_in_seconds: int,
        handles: Callable | list[Callable],
        delay_start: int = 0,
        args: list[any] = None,
    ):
        loop = aio.create_task(
            self.run_loop(interval_in_seconds, handles, delay_start, args)
        )
        self.loops.add(loop)
        loop.add_done_callback(self.loops.remove)

    async def run_loop(
        self,
        interval_in_seconds: int,
        handles: Callable | list[Callable],
        delay_start: int = 0,
        args: list[any] = None,
    ):
        if not isinstance(handles, list):
            handles = [handles]

        await aio.sleep(delay_start)

        while not self.cancel_request.is_set():
            execution_time = 0
            try:
                start_time = time()
                for handle in handles:
                    await handle(*args or [])
                execution_time = time() - start_time
            except InvalidDataException as error:
                logger.warning("{message}, skipping this loop", message=error.message)

            if execution_time < interval_in_seconds:
                await aio.sleep(interval_in_seconds - execution_time)
            else:
                await aio.sleep(interval_in_seconds)



from typing import ClassVar

from requests.exceptions import HTTPError

from solaredge2mqtt.eventbus import BaseEvent, EventBus
from solaredge2mqtt.exceptions import InvalidDataException
from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import OpenWeatherMapOneCall
from solaredge2mqtt.mqtt import MQTTPublishEvent
from solaredge2mqtt.service.http import HTTPClient
from solaredge2mqtt.settings import ServiceSettings

ONECALL_URL = "https://api.openweathermap.org/data/3.0/onecall"
TIMEMACHINE_URL = "https://api.openweathermap.org/data/3.0/onecall/timemachine"


class WeatherUpdateEvent(BaseEvent):
    EVENT_TYPE: ClassVar[str] = "weather_update"
    weather: OpenWeatherMapOneCall

    @classmethod
    async def emit(cls, event_bus: EventBus, **kwargs) -> None:
        await event_bus.emit(cls(**kwargs))
        await MQTTPublishEvent.emit(
            event_bus, topic="weather/current", payload=kwargs["weather"].current
        )


class WeatherClient(HTTPClient):
    def __init__(self, settings: ServiceSettings, event_bus: EventBus) -> None:
        super().__init__("Weather API")

        self.location = settings.location
        self.settings = settings.weather

        self.event_bus = event_bus

    async def loop(self):
        weather = self.get_weather()
        await WeatherUpdateEvent.emit(self.event_bus, weather=weather)

    def get_weather(self) -> OpenWeatherMapOneCall:
        try:
            logger.info("Reading weather data from OpenWeatherMap")
            result = self._get(
                ONECALL_URL,
                params={
                    "lat": self.location.latitude,
                    "lon": self.location.longitude,
                    "exclude": "minutely,daily,alerts",
                    "units": "metric",
                    "lang": self.settings.language,
                    "appid": self.settings.api_key.get_secret_value(),
                },
                timeout=7,
            )

            logger.trace(result)

            if result is None:
                raise InvalidDataException(
                    "Unable to read weather data from OpenWeatherMap"
                )

            weather = OpenWeatherMapOneCall(**result)

            return weather
        except HTTPError as error:
            status_code = error.response.status_code
            if status_code == 401:
                error_msg = (
                    "Invalid OpenWeatherMap API key or no subscription to OneCall-API"
                )
            else:
                error_msg = "Unable to read weather data from OpenWeatherMap (HTTP {status_code})"

            raise InvalidDataException(error_msg) from error



from __future__ import annotations

from asyncio import to_thread
from typing import Callable, ClassVar

from pydantic import BaseModel, Field


class BaseEvent(BaseModel):
    EVENT_TYPE: ClassVar[str] = "unknown"
    to_thread: bool = Field(False)

    @classmethod
    async def emit(cls, event_bus: EventBus, **kwargs) -> None:
        await event_bus.emit(cls(**kwargs))


class EventBus:
    def __init__(self) -> None:
        self._events: dict[str, list[Callable]] = {}

    def subscribe(self, event: type[BaseEvent], callback: Callable) -> None:
        if event.EVENT_TYPE not in self._events:
            self._events[event.EVENT_TYPE] = []
        self._events[event.EVENT_TYPE].append(callback)

    def unsubscribe(self, event: type[BaseEvent], callback: Callable) -> None:
        if event.EVENT_TYPE in self._events:
            self._events[event.EVENT_TYPE].remove(callback)
            if not self._events[event.EVENT_TYPE]:
                self._events.pop(event.EVENT_TYPE)

    async def emit(self, event: BaseEvent) -> None:
        if event.EVENT_TYPE in self._events:
            for callback in self._events[event.EVENT_TYPE]:
                if event.to_thread:
                    await to_thread(callback(event))
                else:
                    await callback(event)



import base64
import hashlib

from pydantic import BaseModel, Field, computed_field


class HomeAssistantBaseModel(BaseModel):
    client_id: str = Field(..., exclude=True)

    def hash_unique_id(self, ids: list[str | int]) -> str:
        ids.append(self.client_id)
        unique_id = "_".join([str(id) for id in ids])

        hash_obj = hashlib.sha256(unique_id.encode())
        hash_digest = hash_obj.digest()

        base64_encoded = base64.urlsafe_b64encode(hash_digest).decode()

        return base64_encoded[:10]


class HomeAssistantDevice(HomeAssistantBaseModel):
    name: str
    state_topic: str = Field(exclude=True)
    manufacturer: str | None = Field(None)
    model: str | None = Field(None)
    hw_version: str | None = Field(None)
    serial_number: str | None = Field(None)
    sw_version: str | None = Field(None)
    via_device: str | None = Field(None)

    @computed_field
    @property
    def identifiers(self) -> str:
        return self.hash_unique_id([self.manufacturer, self.model, self.serial_number])


class HomeAssistantEntity(HomeAssistantBaseModel):
    name: str
    device: HomeAssistantDevice
    state_class: str | None = None
    unit_of_measurement: str | None = None
    jsonpath: str | None = Field(None, exclude=True)

    def __init__(self, device: HomeAssistantDevice, **data):
        super().__init__(client_id=device.client_id, device=device, **data)

    @computed_field
    @property
    def unique_id(self) -> str:
        return self.hash_unique_id(
            [self.device.identifiers, self.name, self.state_topic, self.value_template]
        )

    @computed_field
    @property
    def state_topic(self) -> str:
        return self.device.state_topic

    @computed_field
    @property
    def value_template(self) -> str | None:
        return f"{{{{ value_json.{self.jsonpath} }}}}" if self.jsonpath else None
